from shillelagh.fields import Field
from shillelagh.fields import Order
from shillelagh.filters import Equal


def test_comparison():
    field1 = Field(filters=[Equal], order=Order.ASCENDING, exact=True)
    field2 = Field(filters=[Equal], order=Order.ASCENDING, exact=True)
    field3 = Field(filters=[Equal], order=Order.ASCENDING, exact=False)

    assert field1 == field2
    assert field1 != field3
    assert field1 != 42



from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union


# A value corresponding to a constraint is one of:
#     None
#         This means you have no index for that constraint. SQLite will have to
#         iterate over every row for it.
#     integer
#         This is the argument number for the constraintargs being passed
#         into the Filter() function of your cursor.
#     (integer, boolean)
#         By default SQLite will check what you return. If you set the boolean
#         to False then SQLite won’t do that double checking.
Constraint = Union[None, int, Tuple[int, bool]]

# A row of data
Row = Dict[str, Any]

# An index is a tuple with a column index and an operator to filter it
Index = Tuple[int, int]



import os

import apsw
from shillelagh.adapters.api.weatherapi import WeatherAPI
from shillelagh.backends.apsw.vt import VTModule


def test_weatherapi(requests_mock):
    url = "https://api.weatherapi.com/v1/history.json?key=f426b51ea9aa4e4ab68190907202309&q=94923&dt=2020-10-20"
    payload = {
        "forecast": {
            "forecastday": [
                {
                    "hour": [
                        {"time": "2020-10-20 11:00", "temp_c": 20.0},
                        {"time": "2020-10-20 12:00", "temp_c": 20.1},
                        {"time": "2020-10-20 13:00", "temp_c": 20.2},
                    ],
                },
            ],
        },
    }
    requests_mock.get(url, json=payload)

    connection = apsw.Connection(":memory:")
    cursor = connection.cursor()
    connection.createmodule("weatherapi", VTModule(WeatherAPI))
    cursor.execute(
        f"CREATE VIRTUAL TABLE bodega_bay USING weatherapi(94923, f426b51ea9aa4e4ab68190907202309)",
    )

    sql = "SELECT * FROM bodega_bay WHERE ts = '2020-10-20T12:00:00'"
    data = list(cursor.execute(sql))
    assert data == [(20.1, "2020-10-20T12:00:00")]

    try:
        os.unlink("weatherapi_cache.sqlite")
    except FileNotFoundError:
        pass


def test_weatherapi_api_error(requests_mock):
    url1 = "https://api.weatherapi.com/v1/history.json?key=f426b51ea9aa4e4ab68190907202309&q=94923&dt=2020-10-20"
    payload = {
        "forecast": {
            "forecastday": [
                {
                    "hour": [
                        {"time": "2020-10-20 11:00", "temp_c": 20.0},
                        {"time": "2020-10-20 12:00", "temp_c": 20.1},
                        {"time": "2020-10-20 13:00", "temp_c": 20.2},
                    ],
                },
            ],
        },
    }
    requests_mock.get(url1, json=payload)

    url2 = "https://api.weatherapi.com/v1/history.json?key=f426b51ea9aa4e4ab68190907202309&q=94923&dt=2020-10-21"
    requests_mock.get(url2, status_code=404)

    connection = apsw.Connection(":memory:")
    cursor = connection.cursor()
    connection.createmodule("weatherapi", VTModule(WeatherAPI))
    cursor.execute(
        f"CREATE VIRTUAL TABLE bodega_bay USING weatherapi(94923, f426b51ea9aa4e4ab68190907202309)",
    )

    sql = "SELECT * FROM bodega_bay WHERE ts >= '2020-10-20T12:00:00' AND ts <= '2020-10-21T12:00:00'"
    data = list(cursor.execute(sql))
    assert data == [
        (20.1, "2020-10-20T12:00:00"),
        (20.2, "2020-10-20T13:00:00"),
    ]

    try:
        os.unlink("weatherapi_cache.sqlite")
    except FileNotFoundError:
        pass



from datetime import date
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import Iterator

import dateutil.parser
import requests
import requests_cache
from shillelagh.adapters.base import Adapter
from shillelagh.fields import DateTime
from shillelagh.fields import Float
from shillelagh.fields import Order
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import Row


requests_cache.install_cache(
    cache_name="weatherapi_cache", backend="sqlite", expire_after=180,
)


class WeatherAPI(Adapter):

    ts = DateTime(filters=[Range], order=Order.ASCENDING, exact=False)
    temperature = Float()

    def __init__(self, location: str, api_key: str):
        self.location = location
        self.api_key = api_key

    def get_data(self, bounds: Dict[str, Filter]) -> Iterator[Row]:
        ts_range = bounds["ts"]
        today = date.today()
        start = ts_range.start.date() if ts_range.start else today - timedelta(days=7)
        end = ts_range.end.date() if ts_range.end else today

        while start <= end:
            url = (
                f"https://api.weatherapi.com/v1/history.json?key={self.api_key}"
                f"&q={self.location}&dt={start}"
            )
            response = requests.get(url)
            if response.ok:
                payload = response.json()
                hourly_data = payload["forecast"]["forecastday"][0]["hour"]
                for record in hourly_data:
                    dt = dateutil.parser.parse(record["time"])
                    yield {
                        "rowid": int(dt.timestamp()),
                        "ts": dt.isoformat(),
                        "temperature": record["temp_c"],
                    }

            start += timedelta(days=1)



import inspect
from typing import Dict
from typing import Iterator

from shillelagh.fields import Field
from shillelagh.filters import Filter
from shillelagh.types import Row


class Adapter:
    def get_columns(self) -> Dict[str, Field]:
        return dict(
            inspect.getmembers(self, lambda attribute: isinstance(attribute, Field)),
        )

    def get_data(self, bounds: Dict[str, Filter]) -> Iterator[Row]:
        raise NotImplementedError("Subclasses must implement `get_data`")

    def insert_row(self, row: Row) -> int:
        raise NotImplementedError("Subclasses must implement `insert_row`")

    def delete_row(self, row_id: int) -> None:
        raise NotImplementedError("Subclasses must implement `delete_row`")

    def update_row(self, row_id: int, row: Row) -> None:
        # Subclasses are free to implement inplace updates
        self.delete_row(row_id)
        self.insert_row(row)

    def close(self) -> None:
        pass



from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional

import apsw
import pytest
from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.vt import VTModule
from shillelagh.backends.apsw.vt import VTTable
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import Row


class DummyAdapter(Adapter):

    age = Float(filters=[Range], order=Order.NONE, exact=True)
    name = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    pets = Integer()

    def __init__(self):
        self.data = [
            {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
            {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        ]

    def get_data(self, bounds: Dict[str, Filter]) -> Iterator[Dict[str, Any]]:
        data = self.data[:]

        for column in ["name", "age"]:
            if column in bounds:
                data = [row for row in data if bounds[column].check(row[column])]

        yield from iter(data)

    def insert_row(self, row: Row) -> int:
        row_id: Optional[int] = row["rowid"]
        if row_id is None:
            row["rowid"] = row_id = max(row["rowid"] for row in self.data) + 1

        self.data.append(row)

        return row_id

    def delete_row(self, row_id: int) -> None:
        self.data = [row for row in self.data if row["rowid"] != row_id]


def test_vt_module():
    table = VTTable(DummyAdapter)
    vt_module = VTModule(DummyAdapter)
    create_table, table = vt_module.Create(None, "", "", "table")
    assert (
        create_table
        == """CREATE TABLE "table" ("age" REAL, "name" TEXT, "pets" INTEGER)"""
    )


def test_virtual_best_index():
    table = VTTable(DummyAdapter())
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
        ],
        [(1, False)],  # ORDER BY name ASC
    )
    assert result == (
        [(0, True), None, (1, True)],
        42,
        f"[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}], [0, {apsw.SQLITE_INDEX_CONSTRAINT_LE}]]",
        True,
        666,
    )


def test_virtual_best_index_operator_not_supported():
    table = VTTable(DummyAdapter())
    result = table.BestIndex(
        [(1, apsw.SQLITE_INDEX_CONSTRAINT_MATCH)],  # name LIKE?
        [(1, False)],  # ORDER BY name ASC
    )
    assert result == ([None], 42, "[]", True, 666)


def test_virtual_best_index_no_order_by():
    table = VTTable(DummyAdapter())
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
        ],
        [(0, True)],  # ORDER BY age DESC
    )
    assert result == (
        [(0, True), None, (1, True)],
        42,
        f"[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}], [0, {apsw.SQLITE_INDEX_CONSTRAINT_LE}]]",
        False,
        666,
    )


def test_virtual_disconnect():
    table = VTTable(DummyAdapter())
    table.Disconnect()  # no-op


def test_update_insert_row():
    adapter = DummyAdapter()
    table = VTTable(adapter)

    new_row_id = table.UpdateInsertRow(None, [6, "Charlie", 1])
    assert new_row_id == 2
    assert list(adapter.get_data({})) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
        {"age": 6, "name": "Charlie", "pets": 1, "rowid": 2},
    ]

    new_row_id = table.UpdateInsertRow(4, [40, "Dani", 2])
    assert new_row_id == 4
    assert list(adapter.get_data({})) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
        {"age": 6, "name": "Charlie", "pets": 1, "rowid": 2},
        {"age": 40, "name": "Dani", "pets": 2, "rowid": 4},
    ]


def test_update_delete_row():
    adapter = DummyAdapter()
    table = VTTable(adapter)

    table.UpdateDeleteRow(0)
    assert list(adapter.get_data({})) == [
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
    ]


def test_update_change_row():
    adapter = DummyAdapter()
    table = VTTable(adapter)

    table.UpdateChangeRow(1, 1, [24, "Bob", 4])
    assert list(adapter.get_data({})) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 24, "name": "Bob", "pets": 4, "rowid": 1},
    ]

    table.UpdateChangeRow(1, 2, [24, "Bob", 4])
    assert list(adapter.get_data({})) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 24, "name": "Bob", "pets": 4, "rowid": 2},
    ]


def test_cursor():
    table = VTTable(DummyAdapter())
    cursor = table.Open()
    cursor.Filter(42, "[]", [])
    assert cursor.current_row == (0, 20, "Alice", 0)
    assert cursor.Rowid() == 0
    assert cursor.Column(0) == 20

    cursor.Next()
    assert cursor.current_row == (1, 23, "Bob", 3)

    assert not cursor.Eof()
    cursor.Next()
    assert cursor.Eof()
    cursor.Close()


def test_cursor_with_constraints():
    table = VTTable(DummyAdapter())
    cursor = table.Open()
    cursor.Filter(42, f"[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}]]", ["Alice"])
    assert cursor.current_row == (0, 20, "Alice", 0)

    assert not cursor.Eof()
    cursor.Next()
    assert cursor.Eof()


def test_cursor_with_constraints_invalid_filter():
    table = VTTable(DummyAdapter())
    cursor = table.Open()

    with pytest.raises(Exception) as excinfo:
        cursor.Filter(42, f"[[1, {apsw.SQLITE_INDEX_CONSTRAINT_MATCH}]]", ["Alice"])

    assert str(excinfo.value) == "No valid filter found"



from enum import Enum
from typing import List
from typing import Optional

import dateutil.parser
from shillelagh.filters import Filter


class Order(Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"
    NONE = "none"


class Field:
    def __init__(
        self,
        filters: Optional[List[Filter]] = None,
        order: Order = Order.NONE,
        exact: bool = False,
    ):
        self.filters = filters or []
        self.order = order
        self.exact = exact

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Field):
            return NotImplemented

        return (
            self.filters == other.filters
            and self.order == other.order
            and self.exact == other.exact
        )


class Integer(Field):
    type = "INTEGER"
    parse = int


class Float(Field):
    type = "REAL"
    parse = float


class String(Field):
    type = "TEXT"
    parse = str


class DateTime(Field):
    type = "TIMESTAMP"

    @staticmethod
    def parse(value):
        return dateutil.parser.parse(value)



import apsw
from shillelagh.adapters.file.csvfile import CSVFile
from shillelagh.backends.apsw.vt import VTModule

if __name__ == "__main__":
    connection = apsw.Connection(":memory:")
    cursor = connection.cursor()
    connection.createmodule("csvfile", VTModule(CSVFile))
    cursor.execute(f"CREATE VIRTUAL TABLE test USING csvfile(test.csv)")

    sql = 'SELECT * FROM test WHERE "index" > 11'
    print(sql)
    for row in cursor.execute(sql):
        print(row)
    print("==")

    sql = """INSERT INTO test ("index", temperature, site) VALUES (14, 10.1, 'New_Site')"""
    print(sql)
    cursor.execute(sql)

    sql = 'SELECT * FROM test WHERE "index" > 11'
    print(sql)
    for row in cursor.execute(sql):
        print(row)
    print("==")

    sql = "DELETE FROM test WHERE site = 'New_Site'"
    print(sql)
    cursor.execute(sql)
    sql = "SELECT * FROM test"
    print(sql)
    for row in cursor.execute(sql):
        print(row)
    print("==")
